# -*- coding: utf-8 -*-

from ascii_graph import Pyasciigraph
from dateutil import parser
from elasticsearch_data import LogData, es_identity, es_fields, es_timefield
import pandas as pd


pd.set_option('display.width', 1000)

IDLE_TIME = 3600


def ascii_funnel(info, funnel_data):
    """Return ascii graph by funnel data.

    Parameters
    ----------
    info: string
    funnel_data: list of tuple

    Returns list of ascii string.
    -------

    """
    graph = Pyasciigraph()
    return graph.graph(info, funnel_data)


class Stage(object):
    def __init__(self, total=0, flag=False, name=None):
        self.total = total
        # name -> {"state_name": "index"}
        self.name = name
        self.trend = 0.0
        self.flag = flag
        # flag usually go with last_record_time
        self.last_record_time = None


def is_idle(last_time, current_time,):
    between_time = (parser.parse(current_time) - parser.parse(last_time)).total_seconds()
    if between_time > IDLE_TIME:
        return True
    else:
        return False


class FunnelData(LogData):

    def __init__(self, host=None, port=None, index_name=None, start_time=None, end_time=None, add_query=None):
        LogData.__init__(self, host, port, index_name, start_time, end_time, add_query)

        self.dataframe = None
        self.integrate_data()
        """list: stages is a list.
        Example:
            [Stage(name={"state_name": "index"}), Stage(name={"state_name": explore})]
        """
        self.stages = None

    def integrate_data(self):
        df = []
        size = 50
        pages = self.total / size
        print "Integrating data from Elasticsearch..."
        for i in range(0, pages + 1):
            res = self.result(start=(i * size), size=size)
            for hit in res:
                df.append({field: hit['fields'][field][0] for field in es_fields})
        self.dataframe = pd.DataFrame(df, columns=es_fields).sort(['@timestamp'], ascending=True)

    def set_stages(self, stages):
        """Set funnel stages before you calculating it

        :param stages: a list of dict
        :return: a list of list
        """

        # First is for count, so the initial value is zero.
        # Second is a True or False flag for calculating.
        # Third is a dict of stage name.
        self.stages = [Stage(name=stage) for stage in stages]
        # Init first flag
        (self.stages[0]).flag = True
        return self.stages

    def count_funnel(self):
        if self.stages is None:
            raise ValueError("stages cannot be empty.")

        for browser_id in self.browser_ids:
            # print "Browser ID: " + browser_id

            # Select data set by browser_id
            user_data_set = self.dataframe.loc[self.dataframe[es_identity] == browser_id]

            for index, record in user_data_set.iterrows():
                self._calculate(record)

        return self.stages_percentage()

    def _calculate(self, record):
        """
        Calculate stages count by each record.

        The record input looks like:

            @timestamp          2016-08-24T12:15:10.558791129Z
            sessionid                      3929-45d7-9f53-af87
            action                                state_change
            state                                   searchpage
        """
        # init
        (self.stages[0]).flag = True
        (self.stages[0]).last_record_time = record[es_timefield]

        stage_idx = 0
        while stage_idx < len(self.stages):
            # Search flag using while loop

            # If a stage flag with a switch true
            if (self.stages[stage_idx]).flag:
                stage_key = (self.stages[stage_idx]).name.keys()[0]
                stage_val = (self.stages[stage_idx]).name.values()[0]

                record_stage_val = record[stage_key]

                # Match!!
                if stage_val == record_stage_val:
                    # Check time period for idle gap
                    # Update the current stage info
                    if is_idle((self.stages[stage_idx]).last_record_time, record[es_timefield]) is False:
                        (self.stages[stage_idx]).total += 1
                    (self.stages[stage_idx]).flag = False

                    if (stage_idx + 1) >= len(self.stages):
                        # Restart a traversal
                        (self.stages[0]).flag = True
                        (self.stages[0]).last_record_time = record[es_timefield]
                    else:
                        # Always set for starting the next stage
                        (self.stages[stage_idx + 1]).flag = True
                        (self.stages[stage_idx + 1]).last_record_time = record[es_timefield]
                    break
                else:
                    # Mismatched, set current stage into False,
                    # and set first stage is True for restart a traversal.
                    (self.stages[stage_idx]).flag = False
                    (self.stages[0]).flag = True
                    (self.stages[0]).last_record_time = record[es_timefield]

                    # Although current record stage don't match current stage,
                    # it still can compare with the first stage for traversal.
                    first_stage_val = (self.stages[0]).name.values()[0]

                    # Match!!
                    if first_stage_val == record_stage_val:
                        # Check time period for idle gap
                        # Update the current stage info
                        if is_idle((self.stages[stage_idx]).last_record_time, record[es_timefield]) is False:
                            (self.stages[0]).total += 1
                        (self.stages[0]).flag = False

                        # Always set for starting the next stage
                        (self.stages[1]).flag = True
                        (self.stages[1]).last_record_time = record[es_timefield]
                        break

            stage_idx += 1

    def stages_percentage(self):
        for idx, stage in enumerate(self.stages):
            if idx == 0:
                stage.trend = 100
            else:
                if (self.stages[idx - 1]).total == 0:
                    stage.trend = 0.0
                else:
                    trend = (stage.total * 100.0) / (self.stages[idx - 1]).total
                    stage.trend = trend

        return self.stages



# -*- coding: utf-8 -*-
import json

from elasticsearch_data import es_index, es_port, es_host
from flask import Flask, Response, request
from funnel import FunnelData, ascii_funnel

app = Flask(__name__)


@app.route('/funnel', methods=['POST'])
def handler():
    data = request.data
    data_dict = json.loads(data)
    funnel_data = FunnelData(host=es_host, port=es_port, index_name=es_index,
                             start_time=data_dict['start_time'],
                             end_time=data_dict['end_time'])
    funnel_data.set_stages([{'state_name': 'landingpage'},
                           {'state_name': 'login'},
                           {'state_name': 'searchpage'}])

    fd = funnel_data.count_funnel()
    fd = [('\t' + '{0:.1f}'.format(d[3].values()[0]) + '% \t' + d[2].values()[0], d[0]) for d in fd]
    return Response(' \n'.join(ascii_funnel('Funnel', fd)), mimetype='text/plain')


@app.route('/funnel_list', methods=['POST'])
def funnel_list():
    funnels = ['1. landingpage -> login -> searchpage',
               '2. landingpage -> login -> searchpage -> searchpage -> user',
               '-----------------------------------------------------------------------',
               'An example for send request with time format: 1, 2016-03-06, 2016-03-28']
    return Response(' \n'.join(funnels), mimetype='text/plain')


if __name__ == '__main__':
    app.run('0.0.0.0', 5578, debug=True)


# -*- coding: utf-8 -*-

from elasticsearch import Elasticsearch

import os
import ConfigParser

config = ConfigParser.ConfigParser()
config.read(os.path.expanduser('~/.elastic_funnel'))

es_host = None
es_port = None
es_index = None

if config.has_section('elastic'):
    es_host = os.environ['ELASTIC_HOST'] if os.environ.get('ELASTIC_HOST') else config.get('elastic', 'host')
    es_port = os.environ['ELASTIC_PORT'] if os.environ.get('ELASTIC_PORT') else config.get('elastic', 'port')
    es_index = os.environ['ELASTIC_INDEX'] if os.environ.get('ELASTIC_INDEX') else config.get('elastic', 'index')
    es_query = os.environ['ELASTIC_QUERY'] if os.environ.get('ELASTIC_QUERY') else config.get('elastic', 'query')
    es_identity = os.environ['ELASTIC_IDENTITY'] if os.environ.get('ELASTIC_IDENTITY') else config.get('elastic', 'identity')
    es_timefield = os.environ['ELASTIC_TIMEFIELD'] if os.environ.get('ELASTIC_TIMEFIELD') else config.get('elastic', 'timefield')
    es_stagefield = os.environ['ELASTIC_STAGEFIELD'] if os.environ.get('ELASTIC_STAGEFIELD') else config.get('elastic', 'stagefield')
    es_fields = [es_timefield, es_identity, es_stagefield]
else:
    print "Can't find the elastic section."
    exit()


def search_syntax(start=0, size=0, start_time='1987-03-24T00:00:00', end_time=None, add_query=None):
    add_query = ' AND ' + add_query if add_query else ''
    return {
        "from": start,
        "size": size,
        "fields": es_fields,
        "query": {
            "bool": {
                "must": [
                    {
                        "query_string": {
                            "query": es_query + add_query
                        }
                    },
                    {
                        "range": {
                            es_timefield: {
                                # "gte": "now-" + str(day) + "d",
                                "gte": start_time,
                                "lte": end_time,
                                "time_zone": "+08:00"
                            }
                        }
                    }
                ]
            }
        },
        "sort": [
           {
              es_timefield: {
                 "order": "asc"
              }
           }
        ],
        "aggs": {
            "identity_tag": {
                "terms": {
                    "field": es_identity,
                    "size": 0
                }
            }
        }
    }


class LogData(object):

    def __init__(self, host=None, port=None, index_name=None, start_time=None, end_time=None, add_query=None):
        """LogData is a class to collect log from Elasticsearch.

        Parameters
        ----------
        host: string, Elasticsearch IP
        index_name: string, index name in Elasticsearch
        start_time: date string, format looks like '2016-03-24T00:00:00'
        end_time: date string, format looks like '2016-03-24T00:00:00'
                  if empty this field, 'now' will use.

        Returns
        -------

        """
        if host:
            self.host = host
        else:
            raise ValueError("Elasticsearch host cannot be empty.")

        if port:
            self.port = port
        else:
            self.port = 9200

        if index_name:
            self.index_name = index_name
        else:
            raise ValueError("index name cannot be empty.")

        if start_time:
            self.start_time = start_time
        else:
            self.start_time = '1987-03-24T00:00:00'

        if end_time:
            self.end_time = end_time
        else:
            self.end_time = 'now'
        self.add_query = add_query

        self.es = Elasticsearch([{'host': self.host, 'port': self.port}])
        self.total, self.browser_ids = self._total_log_browser_ids()

    def _total_log_browser_ids(self):
        search_result = self.es.search(index=self.index_name, body=search_syntax(start=0, size=0,
                                                                                 start_time=self.start_time,
                                                                                 end_time=self.end_time,
                                                                                 add_query=self.add_query))
        return search_result['hits']['total'], [bucket['key'] for bucket in search_result['aggregations']['identity_tag']['buckets']]

    def result(self, start=0, size=0):
        search_result = self.es.search(index=self.index_name, body=search_syntax(start=start, size=size,
                                                                                 start_time=self.start_time,
                                                                                 end_time=self.end_time,
                                                                                 add_query=self.add_query))
        return search_result['hits']['hits']


# -*- coding: utf-8 -*-

import argparse

from funnel import FunnelData, ascii_funnel
from elasticsearch_data import es_host, es_port, es_index, es_stagefield


class CliArgs:
    def __init__(self):
        p = argparse.ArgumentParser(prog='elastic_funnel')
        p.add_argument('--host', help='Host of Elasticsearch')
        p.add_argument('--port', type=int, help='Port of Elasticsearch')
        p.add_argument('--index', help='Index name of Elasticsearch, e.g., user-behavior-log-*')
        p.add_argument('--stages', required=True, nargs='+', help='Set a path of stages , '
                                                                  'e.g., landingpage login searchpage')
        p.add_argument('--start', help='Start time of log, e.g., 2016-03-24T00:00:00')
        p.add_argument('--end', help='End time of log, e.g., 2016-03-28T23:59:59')
        p.add_argument('--add_query', help='Additional query using syntax of Lucene, '
                                       'https://lucene.apache.org/core/2_9_4/queryparsersyntax.html. '
                                       'You can narrow you search target by syntax, e.g., country:US')
        args = p.parse_args()

        self.host = args.host if args.host else es_host
        self.port = args.port if args.port else es_port
        self.index = args.index if args.index else es_index
        self.stages = args.stages
        self.start = args.start
        self.end = args.end
        self.add_query = args.add_query


def main():
    cli_args = CliArgs()

    funnel_data = FunnelData(host=cli_args.host, port=cli_args.port, index_name=cli_args.index,
                             start_time=cli_args.start, end_time=cli_args.end, add_query=cli_args.add_query)
    funnel_data.set_stages([{es_stagefield: stage} for stage in cli_args.stages])

    funnel_stages = funnel_data.count_funnel()

    funnel_stages = [('\t' + '{0:.1f}'.format(stage.trend) + '% \t' + stage.name.values()[0], stage.total) for stage in funnel_stages]

    print ' \n'.join([g.encode('utf-8') for g in ascii_funnel('Funnel: ' + ' --> '.join(cli_args.stages), funnel_stages)])


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
import re
from setuptools import setup, find_packages

pkg_file = open("elastic_funnel/__init__.py").read()
metadata = dict(re.findall("__([a-z]+)__\s*=\s*'([^']+)'", pkg_file))
description = open('README.md').read()
requirements = [i.strip() for i in open("requirements.txt").readlines()]

setup(
    name='elastic_funnel',
    description='Analysis tool for funnel visualization with log from Elasticsearch',
    version=metadata['version'],

    # Author details
    author='yuecen',
    author_email='yuecendev+pypi@gmail.com',
    url='https://github.com/yuecen/elastic_funnel',
    long_description=description,
    license='MIT',
    classifiers=[
        'Intended Audience :: Developers',
        'Programming Language :: Python',
    ],
    keywords='elasticsearch, log, funnel, elastic_funnel',
    install_requires=requirements,
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'elastic_funnel = elastic_funnel.cli:main'
        ]
    }
)